open Format
open Unix

type config = {
  results_directory: string;
  include_directory: string;
  lustrec: string;
  gcc: string;
  compcert: string;
  velus: string;
  otawa: string;
}

type report_unit = {
  lustrec_gcc_wcet: int Array.t Array.t;     (* 4 * 4 *)
  lustrec_compcert_wcet: int Array.t;        (* 4 *)
  velus_wcet: int Array.t                    (* 2 *)
}
let empty_unit () = {
  lustrec_gcc_wcet = Array.make_matrix 4 4 (-1);
  lustrec_compcert_wcet = Array.make 4 (-1);
  velus_wcet = Array.make 2 (-1);
}

let pp_report_unit fmt { lustrec_gcc_wcet; lustrec_compcert_wcet; velus_wcet } =
  Array.iter (Array.iter (fprintf fmt "%d ")) lustrec_gcc_wcet;
  Array.iter (fprintf fmt "%d ") lustrec_compcert_wcet;
  Array.iter (fprintf fmt "%d ") velus_wcet

module SMap = Map.Make(String)

let empty_report = SMap.empty

let pp_report fmt =
  fprintf fmt "@[<v>@[<h>\
               #file L-1G0 L-1G1 L-1G2 L-1G3 \
               L1G0 L1G1 L1G2 L1G3 \
               L2G0 L2G1 L2G2 L2G3 \
               L3G0 L3G1 L3G2 L3G3 \
               L-1C L1C L2C L3C \
               Vnf V@]@ %a@]@."
    (fun fmt -> SMap.iter (fun x -> fprintf fmt "@[<h>%s %a@]@ " x pp_report_unit))

let report_file cfg = Filename.concat cfg.results_directory "report"

let write_report cfg report =
  let oc = open_out (report_file cfg) in
  let fmt = formatter_of_out_channel oc in
  pp_report fmt report;
  close_out oc

let config_format_tags =
  set_tags true;
  let mark_open_stag = function
    | String_tag "bold" ->
      "\x1b[1m"
    | String_tag "red" ->
      "\x1b[31m"
    | String_tag "magenta" ->
      "\x1b[35m"
    | String_tag "green" ->
      "\x1b[32m"
    | _ ->
      ""
  in
  let mark_close_stag _ = "\x1b[0m" in
  let stag_fns = get_formatter_stag_functions () in
  set_formatter_stag_functions { stag_fns with mark_open_stag; mark_close_stag }

let pp_header fmt = fprintf fmt "@{<bold>%s:@}"
let err_f = "/tmp/err"

let read_whole_file f =
  let ch = open_in f in
  let s = really_input_string ch (in_channel_length ch) in
  close_in ch;
  s

let get_step f =
  let open Filename in
  let rec chop f =
    try chop (chop_extension f) with _ -> f
  in
  chop (basename f)

let lustrec_params cfg o f =
  [
    "-velus";
    "-d";
    cfg.results_directory;
    "-O";
    (match o with
     | 0 -> "-1"
     | _ -> string_of_int o);
    f;
  ]

let velus_params o f =
  (match o with
   | 0 -> ["-nofusion"]
   | _ -> [])
  @ [
    "-main";
    get_step f;
    f
  ]

let gcc_params cfg o f =
  [
    "-O" ^ string_of_int o;
    "-mfloat-abi=hard";
    "-march=armv7-a";
    "-mfpu=vfpv3-d16";
    "-nostdlib";
    "-specs=nano.specs";
    "-specs=nosys.specs";
    "-I" ^ cfg.include_directory;
    cfg.include_directory ^ "/arrow_noalloc.c";
    "eabi_mock.c";
    f;
    (* "-lgcc" *)
  ]

let compcert_params cfg f =
  [
    "-nostdlib";
    "-fstruct-passing";
    "-I" ^ cfg.include_directory;
    cfg.include_directory ^ "/arrow_noalloc.c";
    "eabi_mock.c";
    f;
  ]

let otawa_params step = [
  "-s";
  "trivial";
  "a.out";
  step
]

let cycles log =
  let open Re.Str in
  let reg = "\\([0-9]+\\) cycles" in
  search_forward (regexp reg) log 0 |> ignore;
  matched_group 1 log |> int_of_string

let has_velus_error log =
  let open Re.Str in
  let reg =
    "Velus output Error:.+"
  in
  try
    search_forward (regexp reg) log 0 |> ignore;
    Some (matched_string log)
  with Not_found -> None

let lustrec_compile cfg o f =
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.lustrec
      (lustrec_params cfg o f)
  in
  let ic = open_process_in cmd in
  match close_process_in ic with
  | WEXITED r -> r = 0
  | _ -> false

let gcc_compile cfg o f =
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.gcc
      (gcc_params cfg o f)
  in
  let ic = open_process_in cmd in
  match close_process_in ic with
  | WEXITED r -> r = 0
  | _ -> false

let compcert_compile cfg f =
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.compcert
      (compcert_params cfg f)
  in
  let ic = open_process_in cmd in
  match close_process_in ic with
  | WEXITED r -> r = 0
  | _ -> false

let velus_compile cfg o f =
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.velus
      (velus_params o f)
  in
  let ic = open_process_in cmd in
  match close_process_in ic with
  | WEXITED 0 ->
    let f_s = Filename.remove_extension f ^ ".s" in
    compcert_compile cfg f_s
  | _ -> false

let pp_warning fmt =
  fprintf fmt "@{<magenta>%s@}"

let pp_error fmt =
  fprintf fmt "@{<red>%s@}"

let warning err_f =
  let log = read_whole_file err_f in
  printf "@ %a" pp_warning log

let error err_f =
  let log = read_whole_file err_f in
  printf "@ %a" pp_error log;
  exit 1

let otawa_wcet update cfg step =
  let cmd =
    Filename.quote_command
      ~stderr:err_f
      cfg.otawa
      (otawa_params step)
  in
  let ic = open_process_in cmd in
  let log = try input_line ic with _ -> "" in
  match close_process_in ic with
  | WEXITED n when n <> 2 ->
    update (cycles log)
  | _ -> error err_f

let compile cfg fs =
  printf "@.%a@." pp_header "Compilation + WCET estimation tests";
  let n = List.length fs in
  let n_f = float_of_int n in
  List.fold_left (fun (k, report) f ->
      let p = float_of_int k *. 100. /. n_f in
      printf "%3.0f%% @[<v 2>%s" p (Filename.remove_extension f);
      let k = k + 1 in
      let report_unit = empty_unit () in

      let b = try
          (* Compile with LustreC *)
          for lustrec_o = 0 to 3 do
            if lustrec_compile cfg lustrec_o f then
              let f' = Filename.remove_extension f ^ ".c" in
              let f_c = Filename.concat cfg.results_directory f' in
              let step = get_step f ^ "_step" in

              (* Compile with GCC *)
              for gcc_o = 0 to 3 do
                if gcc_compile cfg gcc_o f_c then
                  otawa_wcet (fun wcet ->
                      report_unit.lustrec_gcc_wcet.(lustrec_o).(gcc_o) <- wcet)
                    cfg step
                else
                  error err_f
              done;

              (* Compile with CompCert *)
              if compcert_compile cfg f_c then
                (* let step = "fun$step$" ^ get_step f in *)
                otawa_wcet (fun wcet ->
                    report_unit.lustrec_compcert_wcet.(lustrec_o) <- wcet)
                  cfg step
              else
                error err_f

            else begin
              let log = read_whole_file err_f in
              match has_velus_error log with
              | Some m ->
                printf "@ %a" pp_warning m;
                raise Exit
              | None -> printf "@ %a" pp_error log
            end
          done;

          (* Compile with Velus *)
          for velus_o = 0 to 1 do
            let f' = Filename.remove_extension f ^ ".velus.lus" in
            let f = Filename.concat cfg.results_directory f' in
            if velus_compile cfg velus_o f then
              let step = "fun$step$" ^ get_step f in
              otawa_wcet (fun wcet ->
                  report_unit.velus_wcet.(velus_o) <- wcet)
                cfg step
            else
              warning err_f
          done;
          true
        with Exit -> false
      in
      printf "@]@.";
      let report =
        if b then
          SMap.add Filename.(basename (remove_extension f)) report_unit report
        else report
      in
     write_report cfg report;
     k, report)
    (0, empty_report) fs

      (* let n = List.length fs in *)
  (* let n_f = float_of_int n in *)
  (* let report, ok, ko, ninits, _ = *)
  (*   List.fold_left *)
  (*     (fun (report, ok, ko, ninits, i) f -> *)
  (*       let i' = i + 1 in *)
  (*       let p = float_of_int i' *. 100. /. n_f in *)
  (*       let success r : _ * _ * _ * _ * _ format * _ = *)
  (*         r, ok + 1, ko, ninits, "@{<green>%s@}", "OK" *)
  (*       in *)
  (*       let ninit r : _ * _ * _ * _ * _ format * _ = *)
  (*         r, ok, ko, ninits + 1, "@{<magenta>%s@}", "NI" *)
  (*       in *)
  (*       let fail r log : _ * _ * _ * _ * _ format * _ = *)
  (*         r, ok, ko + 1, ninits, "@{<red>%s@}", "KO\n" ^ log *)
  (*       in *)
  (*       let report, ok, ko, ninits, fmt_str, check = *)
  (*         let cmd = *)
  (*             Filename.quote_command *)
  (*               ~stderr:err_f *)
  (*               lustrec *)
  (*               (lustrec_params cfg f) *)
  (*           in *)
  (*           let ic = open_process_in cmd in *)
  (*           let b = *)
  (*             match close_process_in ic with WEXITED r -> r = 0 | _ -> false *)
  (*           in *)
  (*           let log = read_whole_file err_f in *)
  (*           if b then *)
  (*             if has_warning log then ninit (add_ninit report f) *)
  (*             else success (add_compiled report f) *)
  (*           else fail report log *)
  (*       in *)
  (*       print_result p f fmt_str check; *)
  (*       write_report cfg report; *)
  (*       report, ok, ko, ninits, i') *)
  (*     (report, 0, 0, 0, 0) *)
  (*     fs *)
  (* in *)
  (* print_results ok ko ninits n; *)
  (* report *)

let expand_path path =
  let home = Unix.getenv "HOME" in
  let open Re.Str in
  global_replace (regexp "~") home path

let set_config () =
  let root = Sys.argv.(1) in
  let lustrec = Sys.argv.(2) in
  let tests_directory = Sys.getcwd () |> Filename.basename in
  let result_dir = root ^ "/../../results/" in
  (try Unix.mkdir result_dir 0o755 with _ -> ());
  let results_directory = result_dir ^ tests_directory in
  (try Unix.mkdir results_directory 0o755 with _ -> ());
  let include_directory = root ^ "/include" in
  let gcc = expand_path "arm-none-eabi-gcc" in
  let otawa = expand_path "owcet" in
  let velus = expand_path "~/velus/velus" in
  let compcert = Filename.(concat (dirname velus) "CompCert/ccomp") in
  { results_directory; include_directory; lustrec; gcc; compcert; velus; otawa; }

let () =
  let cfg = set_config () in
  let lus_fs = Sys.argv.(3) |> Re.Str.(split (regexp_string " ")) in
  (* let report = parse_report cfg in *)
  (* print_endline (Sys.getcwd ()  ); exit 1 |> ignore; *)
  let report = compile cfg lus_fs in
  ignore report
  (* let ignored_fs = *)
  (*   ignored_fs @ List.map Filename.chop_extension (S.elements report.ninits) *)
  (* in *)
  (* print_ignored (List.map (fun f -> f ^ ".c") ignored_fs); *)
  (* let lus_fs = *)
  (*   List.( *)
  (*     sort_uniq *)
  (*       Stdlib.compare *)
  (*       (filter *)
  (*          (fun f -> not (mem (Filename.chop_extension f) ignored_fs)) *)
  (*          lus_fs)) *)
  (* in *)
  (* verify cfg report (next_timeout cfg report) lus_fs *)
